@@ -30,21 +30,47 @@ from ._internal import (
     Config,
     DataFrame,
     SessionContext,
-    Expression,
+    SessionConfig,
+    RuntimeConfig,
     ScalarUDF,
 )
 
+from .common import (
+    DFField,
+    DFSchema,
+)
+
+from .expr import (
+    Analyze,
+    Expr,
+    Filter,
+    Limit,
+    Projection,
+    Sort,
+    TableScan,
+)
+
 __version__ = importlib_metadata.version(__name__)
 
 __all__ = [
     "Config",
     "DataFrame",
     "SessionContext",
-    "Expression",
+    "SessionConfig",
+    "RuntimeConfig",
+    "Expr",
     "AggregateUDF",
     "ScalarUDF",
     "column",
     "literal",
+    "TableScan",
+    "Projection",
+    "DFSchema",
+    "DFField",
+    "Analyze",
+    "Sort",
+    "Limit",
+    "Filter",
 ]
 
 
@@ -67,7 +93,7 @@ class Accumulator(metaclass=ABCMeta):
 
 
 def column(value):
-    return Expression.column(value)
+    return Expr.column(value)
 
 
 col = column
@@ -76,7 +102,7 @@ col = column
 def literal(value):
     if not isinstance(value, pa.Scalar):
         value = pa.scalar(value)
-    return Expression.literal(value)
+    return Expr.literal(value)
 
 
 lit = literal
