@@ -25,95 +25,209 @@ use uuid::Uuid;
 use pyo3::exceptions::{PyKeyError, PyValueError};
 use pyo3::prelude::*;
 
-use parking_lot::RwLock;
-
 use crate::catalog::{PyCatalog, PyTable};
 use crate::dataframe::PyDataFrame;
 use crate::dataset::Dataset;
-use crate::errors::DataFusionError;
+use crate::errors::{py_datafusion_err, DataFusionError};
+use crate::physical_plan::PyExecutionPlan;
+use crate::record_batch::PyRecordBatchStream;
+use crate::sql::logical::PyLogicalPlan;
 use crate::store::StorageContexts;
 use crate::udaf::PyAggregateUDF;
 use crate::udf::PyScalarUDF;
 use crate::utils::wait_for_future;
-use datafusion::arrow::datatypes::Schema;
+use datafusion::arrow::datatypes::{DataType, Schema};
 use datafusion::arrow::pyarrow::PyArrowType;
 use datafusion::arrow::record_batch::RecordBatch;
-use datafusion::config::ConfigOptions;
 use datafusion::datasource::datasource::TableProvider;
 use datafusion::datasource::MemTable;
-use datafusion::execution::context::{SessionConfig, SessionContext};
-use datafusion::prelude::{AvroReadOptions, CsvReadOptions, NdJsonReadOptions, ParquetReadOptions};
+use datafusion::execution::context::{SessionConfig, SessionContext, TaskContext};
+use datafusion::execution::disk_manager::DiskManagerConfig;
+use datafusion::execution::memory_pool::{FairSpillPool, GreedyMemoryPool, UnboundedMemoryPool};
+use datafusion::execution::runtime_env::{RuntimeConfig, RuntimeEnv};
+use datafusion::physical_plan::SendableRecordBatchStream;
+use datafusion::prelude::{
+    AvroReadOptions, CsvReadOptions, DataFrame, NdJsonReadOptions, ParquetReadOptions,
+};
+use datafusion_common::ScalarValue;
+use tokio::runtime::Runtime;
+use tokio::task::JoinHandle;
+
+#[pyclass(name = "SessionConfig", module = "datafusion", subclass, unsendable)]
+#[derive(Clone, Default)]
+pub(crate) struct PySessionConfig {
+    pub(crate) config: SessionConfig,
+}
+
+impl From<SessionConfig> for PySessionConfig {
+    fn from(config: SessionConfig) -> Self {
+        Self { config }
+    }
+}
+
+#[pymethods]
+impl PySessionConfig {
+    #[pyo3(signature = (config_options=None))]
+    #[new]
+    fn new(config_options: Option<HashMap<String, String>>) -> Self {
+        let mut config = SessionConfig::new();
+        if let Some(hash_map) = config_options {
+            for (k, v) in &hash_map {
+                config = config.set(k, ScalarValue::Utf8(Some(v.clone())));
+            }
+        }
+
+        Self { config }
+    }
+
+    fn with_create_default_catalog_and_schema(&self, enabled: bool) -> Self {
+        Self::from(
+            self.config
+                .clone()
+                .with_create_default_catalog_and_schema(enabled),
+        )
+    }
+
+    fn with_default_catalog_and_schema(&self, catalog: &str, schema: &str) -> Self {
+        Self::from(
+            self.config
+                .clone()
+                .with_default_catalog_and_schema(catalog, schema),
+        )
+    }
+
+    fn with_information_schema(&self, enabled: bool) -> Self {
+        Self::from(self.config.clone().with_information_schema(enabled))
+    }
+
+    fn with_batch_size(&self, batch_size: usize) -> Self {
+        Self::from(self.config.clone().with_batch_size(batch_size))
+    }
+
+    fn with_target_partitions(&self, target_partitions: usize) -> Self {
+        Self::from(
+            self.config
+                .clone()
+                .with_target_partitions(target_partitions),
+        )
+    }
+
+    fn with_repartition_aggregations(&self, enabled: bool) -> Self {
+        Self::from(self.config.clone().with_repartition_aggregations(enabled))
+    }
+
+    fn with_repartition_joins(&self, enabled: bool) -> Self {
+        Self::from(self.config.clone().with_repartition_joins(enabled))
+    }
+
+    fn with_repartition_windows(&self, enabled: bool) -> Self {
+        Self::from(self.config.clone().with_repartition_windows(enabled))
+    }
+
+    fn with_repartition_sorts(&self, enabled: bool) -> Self {
+        Self::from(self.config.clone().with_repartition_sorts(enabled))
+    }
+
+    fn with_repartition_file_scans(&self, enabled: bool) -> Self {
+        Self::from(self.config.clone().with_repartition_file_scans(enabled))
+    }
+
+    fn with_repartition_file_min_size(&self, size: usize) -> Self {
+        Self::from(self.config.clone().with_repartition_file_min_size(size))
+    }
+
+    fn with_parquet_pruning(&self, enabled: bool) -> Self {
+        Self::from(self.config.clone().with_parquet_pruning(enabled))
+    }
+}
+
+#[pyclass(name = "RuntimeConfig", module = "datafusion", subclass, unsendable)]
+#[derive(Clone)]
+pub(crate) struct PyRuntimeConfig {
+    pub(crate) config: RuntimeConfig,
+}
+
+#[pymethods]
+impl PyRuntimeConfig {
+    #[new]
+    fn new() -> Self {
+        Self {
+            config: RuntimeConfig::default(),
+        }
+    }
+
+    fn with_disk_manager_disabled(&self) -> Self {
+        let config = self.config.clone();
+        let config = config.with_disk_manager(DiskManagerConfig::Disabled);
+        Self { config }
+    }
+
+    fn with_disk_manager_os(&self) -> Self {
+        let config = self.config.clone();
+        let config = config.with_disk_manager(DiskManagerConfig::NewOs);
+        Self { config }
+    }
+
+    fn with_disk_manager_specified(&self, paths: Vec<String>) -> Self {
+        let config = self.config.clone();
+        let paths = paths.iter().map(|s| s.into()).collect();
+        let config = config.with_disk_manager(DiskManagerConfig::NewSpecified(paths));
+        Self { config }
+    }
+
+    fn with_unbounded_memory_pool(&self) -> Self {
+        let config = self.config.clone();
+        let config = config.with_memory_pool(Arc::new(UnboundedMemoryPool::default()));
+        Self { config }
+    }
+
+    fn with_fair_spill_pool(&self, size: usize) -> Self {
+        let config = self.config.clone();
+        let config = config.with_memory_pool(Arc::new(FairSpillPool::new(size)));
+        Self { config }
+    }
+
+    fn with_greedy_memory_pool(&self, size: usize) -> Self {
+        let config = self.config.clone();
+        let config = config.with_memory_pool(Arc::new(GreedyMemoryPool::new(size)));
+        Self { config }
+    }
+
+    fn with_temp_file_path(&self, path: &str) -> Self {
+        let config = self.config.clone();
+        let config = config.with_temp_file_path(path);
+        Self { config }
+    }
+}
 
 /// `PySessionContext` is able to plan and execute DataFusion plans.
 /// It has a powerful optimizer, a physical planner for local execution, and a
 /// multi-threaded execution engine to perform the execution.
 #[pyclass(name = "SessionContext", module = "datafusion", subclass, unsendable)]
+#[derive(Clone)]
 pub(crate) struct PySessionContext {
-    ctx: SessionContext,
+    pub(crate) ctx: SessionContext,
 }
 
 #[pymethods]
 impl PySessionContext {
-    #[allow(clippy::too_many_arguments)]
-    #[args(
-        default_catalog = "\"datafusion\"",
-        default_schema = "\"public\"",
-        create_default_catalog_and_schema = "true",
-        information_schema = "false",
-        repartition_joins = "true",
-        repartition_aggregations = "true",
-        repartition_windows = "true",
-        parquet_pruning = "true",
-        target_partitions = "None",
-        config_options = "None"
-    )]
+    #[pyo3(signature = (config=None, runtime=None))]
     #[new]
-    fn new(
-        default_catalog: &str,
-        default_schema: &str,
-        create_default_catalog_and_schema: bool,
-        information_schema: bool,
-        repartition_joins: bool,
-        repartition_aggregations: bool,
-        repartition_windows: bool,
-        parquet_pruning: bool,
-        target_partitions: Option<usize>,
-        config_options: Option<HashMap<String, String>>,
-    ) -> Self {
-        let mut options = ConfigOptions::from_env();
-        if let Some(hash_map) = config_options {
-            for (k, v) in &hash_map {
-                if let Ok(v) = v.parse::<bool>() {
-                    options.set_bool(k, v);
-                } else if let Ok(v) = v.parse::<u64>() {
-                    options.set_u64(k, v);
-                } else {
-                    options.set_string(k, v);
-                }
-            }
-        }
-        let config_options = Arc::new(RwLock::new(options));
-
-        let mut cfg = SessionConfig::new()
-            .create_default_catalog_and_schema(create_default_catalog_and_schema)
-            .with_default_catalog_and_schema(default_catalog, default_schema)
-            .with_information_schema(information_schema)
-            .with_repartition_joins(repartition_joins)
-            .with_repartition_aggregations(repartition_aggregations)
-            .with_repartition_windows(repartition_windows)
-            .with_parquet_pruning(parquet_pruning);
-
-        // TODO we should add a `with_config_options` to `SessionConfig`
-        cfg.config_options = config_options;
-
-        let cfg_full = match target_partitions {
-            None => cfg,
-            Some(x) => cfg.with_target_partitions(x),
+    fn new(config: Option<PySessionConfig>, runtime: Option<PyRuntimeConfig>) -> PyResult<Self> {
+        let config = if let Some(c) = config {
+            c.config
+        } else {
+            SessionConfig::default()
         };
-
-        PySessionContext {
-            ctx: SessionContext::with_config(cfg_full),
-        }
+        let runtime_config = if let Some(c) = runtime {
+            c.config
+        } else {
+            RuntimeConfig::default()
+        };
+        let runtime = Arc::new(RuntimeEnv::new(runtime_config)?);
+        Ok(PySessionContext {
+            ctx: SessionContext::with_config_rt(config, runtime),
+        })
     }
 
     /// Register a an object store with the given name
@@ -161,6 +275,7 @@ impl PySessionContext {
     fn create_dataframe(
         &mut self,
         partitions: PyArrowType<Vec<Vec<RecordBatch>>>,
+        py: Python,
     ) -> PyResult<PyDataFrame> {
         let schema = partitions.0[0][0].schema();
         let table = MemTable::try_new(schema, partitions.0).map_err(DataFusionError::from)?;
@@ -169,18 +284,24 @@ impl PySessionContext {
         // table name cannot start with numeric digit
         let name = "c".to_owned()
             + Uuid::new_v4()
-                .to_simple()
+                .simple()
                 .encode_lower(&mut Uuid::encode_buffer());
 
         self.ctx
             .register_table(&*name, Arc::new(table))
             .map_err(DataFusionError::from)?;
-        let table = self.ctx.table(&*name).map_err(DataFusionError::from)?;
+
+        let table = wait_for_future(py, self._table(&name)).map_err(DataFusionError::from)?;
 
         let df = PyDataFrame::new(table);
         Ok(df)
     }
 
+    /// Create a DataFrame from an existing logical plan
+    fn create_dataframe_from_logical_plan(&mut self, plan: PyLogicalPlan) -> PyDataFrame {
+        PyDataFrame::new(DataFrame::new(self.ctx.state(), plan.plan.as_ref().clone()))
+    }
+
     fn register_table(&mut self, name: &str, table: &PyTable) -> PyResult<()> {
         self.ctx
             .register_table(name, table.table())
@@ -209,22 +330,20 @@ impl PySessionContext {
     }
 
     #[allow(clippy::too_many_arguments)]
-    #[args(
-        table_partition_cols = "vec![]",
-        parquet_pruning = "true",
-        file_extension = "\".parquet\""
-    )]
+    #[pyo3(signature = (name, path, table_partition_cols=vec![],
+                        parquet_pruning=true,
+                        file_extension=".parquet"))]
     fn register_parquet(
         &mut self,
         name: &str,
         path: &str,
-        table_partition_cols: Vec<String>,
+        table_partition_cols: Vec<(String, String)>,
         parquet_pruning: bool,
         file_extension: &str,
         py: Python,
     ) -> PyResult<()> {
         let mut options = ParquetReadOptions::default()
-            .table_partition_cols(table_partition_cols)
+            .table_partition_cols(convert_table_partition_cols(table_partition_cols)?)
             .parquet_pruning(parquet_pruning);
         options.file_extension = file_extension;
         let result = self.ctx.register_parquet(name, path, options);
@@ -233,13 +352,13 @@ impl PySessionContext {
     }
 
     #[allow(clippy::too_many_arguments)]
-    #[args(
-        schema = "None",
-        has_header = "true",
-        delimiter = "\",\"",
-        schema_infer_max_records = "1000",
-        file_extension = "\".csv\""
-    )]
+    #[pyo3(signature = (name,
+                        path,
+                        schema=None,
+                        has_header=true,
+                        delimiter=",",
+                        schema_infer_max_records=1000,
+                        file_extension=".csv"))]
     fn register_csv(
         &mut self,
         name: &str,
@@ -295,13 +414,13 @@ impl PySessionContext {
         Ok(())
     }
 
-    #[args(name = "\"datafusion\"")]
+    #[pyo3(signature = (name="datafusion"))]
     fn catalog(&self, name: &str) -> PyResult<PyCatalog> {
         match self.ctx.catalog(name) {
             Some(catalog) => Ok(PyCatalog::new(catalog)),
             None => Err(PyKeyError::new_err(format!(
                 "Catalog with name {} doesn't exist.",
-                &name
+                &name,
             ))),
         }
     }
@@ -311,8 +430,9 @@ impl PySessionContext {
         self.ctx.tables().unwrap()
     }
 
-    fn table(&self, name: &str) -> PyResult<PyDataFrame> {
-        Ok(PyDataFrame::new(self.ctx.table(name)?))
+    fn table(&self, name: &str, py: Python) -> PyResult<PyDataFrame> {
+        let x = wait_for_future(py, self.ctx.table(name)).map_err(DataFusionError::from)?;
+        Ok(PyDataFrame::new(x))
     }
 
     fn table_exist(&self, name: &str) -> PyResult<bool> {
@@ -328,44 +448,43 @@ impl PySessionContext {
     }
 
     #[allow(clippy::too_many_arguments)]
-    #[args(
-        schema = "None",
-        schema_infer_max_records = "1000",
-        file_extension = "\".json\"",
-        table_partition_cols = "vec![]"
-    )]
+    #[pyo3(signature = (path, schema=None, schema_infer_max_records=1000, file_extension=".json", table_partition_cols=vec![]))]
     fn read_json(
         &mut self,
         path: PathBuf,
         schema: Option<PyArrowType<Schema>>,
         schema_infer_max_records: usize,
         file_extension: &str,
-        table_partition_cols: Vec<String>,
+        table_partition_cols: Vec<(String, String)>,
         py: Python,
     ) -> PyResult<PyDataFrame> {
         let path = path
             .to_str()
             .ok_or_else(|| PyValueError::new_err("Unable to convert path to a string"))?;
-
-        let mut options = NdJsonReadOptions::default().table_partition_cols(table_partition_cols);
-        options.schema = schema.map(|s| Arc::new(s.0));
+        let mut options = NdJsonReadOptions::default()
+            .table_partition_cols(convert_table_partition_cols(table_partition_cols)?);
         options.schema_infer_max_records = schema_infer_max_records;
         options.file_extension = file_extension;
-
-        let result = self.ctx.read_json(path, options);
-        let df = wait_for_future(py, result).map_err(DataFusionError::from)?;
+        let df = if let Some(schema) = schema {
+            options.schema = Some(&schema.0);
+            let result = self.ctx.read_json(path, options);
+            wait_for_future(py, result).map_err(DataFusionError::from)?
+        } else {
+            let result = self.ctx.read_json(path, options);
+            wait_for_future(py, result).map_err(DataFusionError::from)?
+        };
         Ok(PyDataFrame::new(df))
     }
 
     #[allow(clippy::too_many_arguments)]
-    #[args(
-        schema = "None",
-        has_header = "true",
-        delimiter = "\",\"",
-        schema_infer_max_records = "1000",
-        file_extension = "\".csv\"",
-        table_partition_cols = "vec![]"
-    )]
+    #[pyo3(signature = (
+        path,
+        schema=None,
+        has_header=true,
+        delimiter=",",
+        schema_infer_max_records=1000,
+        file_extension=".csv",
+        table_partition_cols=vec![]))]
     fn read_csv(
         &self,
         path: PathBuf,
@@ -374,7 +493,7 @@ impl PySessionContext {
         delimiter: &str,
         schema_infer_max_records: usize,
         file_extension: &str,
-        table_partition_cols: Vec<String>,
+        table_partition_cols: Vec<(String, String)>,
         py: Python,
     ) -> PyResult<PyDataFrame> {
         let path = path
@@ -393,7 +512,7 @@ impl PySessionContext {
             .delimiter(delimiter[0])
             .schema_infer_max_records(schema_infer_max_records)
             .file_extension(file_extension)
-            .table_partition_cols(table_partition_cols);
+            .table_partition_cols(convert_table_partition_cols(table_partition_cols)?);
 
         if let Some(py_schema) = schema {
             options.schema = Some(&py_schema.0);
@@ -408,23 +527,23 @@ impl PySessionContext {
     }
 
     #[allow(clippy::too_many_arguments)]
-    #[args(
-        parquet_pruning = "true",
-        file_extension = "\".parquet\"",
-        table_partition_cols = "vec![]",
-        skip_metadata = "true"
-    )]
+    #[pyo3(signature = (
+        path,
+        table_partition_cols=vec![],
+        parquet_pruning=true,
+        file_extension=".parquet",
+        skip_metadata=true))]
     fn read_parquet(
         &self,
         path: &str,
-        table_partition_cols: Vec<String>,
+        table_partition_cols: Vec<(String, String)>,
         parquet_pruning: bool,
         file_extension: &str,
         skip_metadata: bool,
         py: Python,
     ) -> PyResult<PyDataFrame> {
         let mut options = ParquetReadOptions::default()
-            .table_partition_cols(table_partition_cols)
+            .table_partition_cols(convert_table_partition_cols(table_partition_cols)?)
             .parquet_pruning(parquet_pruning)
             .skip_metadata(skip_metadata);
         options.file_extension = file_extension;
@@ -435,25 +554,90 @@ impl PySessionContext {
     }
 
     #[allow(clippy::too_many_arguments)]
-    #[args(
-        schema = "None",
-        file_extension = "\".avro\"",
-        table_partition_cols = "vec![]"
-    )]
+    #[pyo3(signature = (path, schema=None, table_partition_cols=vec![], file_extension=".avro"))]
     fn read_avro(
         &self,
         path: &str,
         schema: Option<PyArrowType<Schema>>,
-        table_partition_cols: Vec<String>,
+        table_partition_cols: Vec<(String, String)>,
         file_extension: &str,
         py: Python,
     ) -> PyResult<PyDataFrame> {
-        let mut options = AvroReadOptions::default().table_partition_cols(table_partition_cols);
+        let mut options = AvroReadOptions::default()
+            .table_partition_cols(convert_table_partition_cols(table_partition_cols)?);
         options.file_extension = file_extension;
-        options.schema = schema.map(|s| Arc::new(s.0));
+        let df = if let Some(schema) = schema {
+            options.schema = Some(&schema.0);
+            let read_future = self.ctx.read_avro(path, options);
+            wait_for_future(py, read_future).map_err(DataFusionError::from)?
+        } else {
+            let read_future = self.ctx.read_avro(path, options);
+            wait_for_future(py, read_future).map_err(DataFusionError::from)?
+        };
+        Ok(PyDataFrame::new(df))
+    }
 
-        let result = self.ctx.read_avro(path, options);
-        let df = PyDataFrame::new(wait_for_future(py, result).map_err(DataFusionError::from)?);
-        Ok(df)
+    fn __repr__(&self) -> PyResult<String> {
+        let id = self.session_id();
+        match id {
+            Ok(value) => Ok(format!("SessionContext(session_id={value})")),
+            Err(err) => Ok(format!("Error: {:?}", err.to_string())),
+        }
+    }
+
+    /// Execute a partition of an execution plan and return a stream of record batches
+    pub fn execute(
+        &self,
+        plan: PyExecutionPlan,
+        part: usize,
+        py: Python,
+    ) -> PyResult<PyRecordBatchStream> {
+        let ctx = Arc::new(TaskContext::new(
+            "task_id".to_string(),
+            "session_id".to_string(),
+            HashMap::new(),
+            HashMap::new(),
+            HashMap::new(),
+            Arc::new(RuntimeEnv::default()),
+        ));
+        // create a Tokio runtime to run the async code
+        let rt = Runtime::new().unwrap();
+        let plan = plan.plan.clone();
+        let fut: JoinHandle<datafusion_common::Result<SendableRecordBatchStream>> =
+            rt.spawn(async move { plan.execute(part, ctx) });
+        let stream = wait_for_future(py, fut).map_err(|e| py_datafusion_err(e))?;
+        Ok(PyRecordBatchStream::new(stream?))
+    }
+}
+
+impl PySessionContext {
+    async fn _table(&self, name: &str) -> datafusion_common::Result<DataFrame> {
+        self.ctx.table(name).await
+    }
+}
+
+fn convert_table_partition_cols(
+    table_partition_cols: Vec<(String, String)>,
+) -> Result<Vec<(String, DataType)>, DataFusionError> {
+    table_partition_cols
+        .into_iter()
+        .map(|(name, ty)| match ty.as_str() {
+            "string" => Ok((name, DataType::Utf8)),
+            _ => Err(DataFusionError::Common(format!(
+                "Unsupported data type '{ty}' for partition column"
+            ))),
+        })
+        .collect::<Result<Vec<_>, _>>()
+}
+
+impl From<PySessionContext> for SessionContext {
+    fn from(ctx: PySessionContext) -> SessionContext {
+        ctx.ctx
+    }
+}
+
+impl From<SessionContext> for PySessionContext {
+    fn from(ctx: SessionContext) -> PySessionContext {
+        PySessionContext { ctx }
     }
 }
